import os

# TODO: Add switch?
plat_map = {
    "windows": "pc",
    "linux": "pc",
    "mac": "macos",
}


def get_files(fpath: os.PathLike) -> list[str]:
    file_list = []
    for root, _, files in os.walk(fpath):
        for file in files:
            file_list.append(os.path.join(root, file))
    return file_list



import os
from contextlib import contextmanager
from tempfile import NamedTemporaryFile

from hgpaktool.utils import padding, parse_manifest, roundup


def test_roundup():
    assert roundup(0) == 0
    assert roundup(1) == 0x10
    assert roundup(0xF) == 0x10
    assert roundup(0x10) == 0x10
    assert roundup(0x11) == 0x20


def test_padding():
    assert padding(0) == 0
    assert padding(1) == 0xF
    assert padding(0xF) == 1
    assert padding(0x10) == 0
    assert padding(0x11) == 0xF


@contextmanager
def tempfile_with_data(data: str):
    """Simple wrapper around the NamedTemporaryFile to write it with data but allow the file to be read."""
    tmp = NamedTemporaryFile("w", delete=False)
    tmp.write(data)
    tmp.close()
    yield tmp
    os.unlink(tmp.name)


def test_parse_manifest():
    # Posix path
    with tempfile_with_data("test/path.mbin\r\n") as tmp:
        assert parse_manifest(tmp.name) == ["test/path.mbin"]
    with tempfile_with_data("test/path.mbin\r\ntest/path2.mbin\r\n") as tmp:
        assert parse_manifest(tmp.name) == ["test/path.mbin", "test/path2.mbin"]
    # Windows path
    with tempfile_with_data("test\\path.mbin\r\n") as tmp:
        assert parse_manifest(tmp.name) == ["test/path.mbin"]
    with tempfile_with_data("test\\path.mbin\r\ntest\\path2.mbin\r\n") as tmp:
        assert parse_manifest(tmp.name) == ["test/path.mbin", "test/path2.mbin"]
    # Mix
    with tempfile_with_data("test/path.mbin\r\ntest\\path2.mbin\r\n") as tmp:
        assert parse_manifest(tmp.name) == ["test/path.mbin", "test/path2.mbin"]



import os.path as op
import shutil
from io import BytesIO
from pathlib import Path
from typing import Literal

import pytest
from utils import get_files, plat_map

from hgpaktool import HGPAKFile
from hgpaktool.api import InvalidFileException
from hgpaktool.utils import normalise_path, parse_manifest

DATA_DIR = op.join(op.dirname(__file__), "data")


@pytest.mark.parametrize("platform", ("windows", "mac", "linux"))
def test_unpack(tmp_path: Path, platform: Literal["windows", "mac", "linux"]):
    with HGPAKFile(op.join(DATA_DIR, f"NMSARC.MeshPlanetSKY.{platform}.pak"), platform) as pak:
        print(pak.filenames)
        assert len(pak.filenames) == 6
        # Extract the files to a temporary directory and analyse it
        pak.unpack(tmp_path)
        assert len(get_files(tmp_path)) == 6
        shutil.rmtree(tmp_path / "models")
        pak.unpack(tmp_path, upper=True)
        files = get_files(tmp_path)
        assert len(files) == 6
        # For each file ensure the base path component up to the extraction path has its capitalization
        # retained. Then check the filename itself is uppercase.
        for fpath in files:
            assert fpath.startswith(str(tmp_path))
            final_path = Path(fpath).relative_to(tmp_path)
            assert str(final_path).upper() == str(final_path)


@pytest.mark.parametrize("platform", ("windows", "mac", "linux"))
def test_unpack_with_manifest(tmp_path: Path, platform: Literal["windows", "mac", "linux"]):
    with HGPAKFile(op.join(DATA_DIR, f"NMSARC.MeshPlanetSKY.{platform}.pak"), platform) as pak:
        assert len(pak.filenames) == 6
        # Extract the files to a temporary directory and analyse it
        pak.unpack(tmp_path, write_manifest=True)
        files = get_files(tmp_path)
        assert len(files) == 7
        # Find the manifest file
        manifest_fpath = None
        for fpath in files:
            if fpath.endswith(".manifest"):
                manifest_fpath = fpath
        assert manifest_fpath is not None
        assert op.exists(manifest_fpath)

        manifest_contents = parse_manifest(manifest_fpath)
        # Remove the base path and normalise the paths of the real files and check that they match the
        # manifest
        norm_paths = []
        for fpath in files:
            if fpath != manifest_fpath:
                norm_paths.append(normalise_path(op.relpath(fpath, tmp_path)))
        assert len(norm_paths) == 6
        assert set(norm_paths) == set(manifest_contents)


@pytest.mark.parametrize("platform", ("windows", "mac", "linux"))
def test_filtered_extraction(platform: Literal["windows", "mac", "linux"]):
    with HGPAKFile(op.join(DATA_DIR, f"NMSARC.MeshPlanetSKY.{platform}.pak"), platform) as pak:
        assert len([x for x in pak.extract("*rainbowplane*")]) == 2
        assert len([x for x in pak.extract("*RAINBOWPLANE*")]) == 2
        assert len([x for x in pak.extract(["*rainbowplane*", "*skycube*"])]) == 4
        assert (
            len([x for x in pak.extract(f"models/planets/sky/skysphere.geometry.mbin.{plat_map[platform]}")])
            == 1
        )
        assert (
            len(
                [
                    x
                    for x in pak.extract(
                        f"MODELS/PLANETS/SKY/SKYSPHERE.GEOMETRY.MBIN.{plat_map[platform].upper()}"
                    )
                ]
            )
            == 1
        )


@pytest.mark.parametrize("platform", ("windows", "mac", "linux"))
@pytest.mark.parametrize("as_buffer", (True, False))
def test_extract_specific_single_file(platform: Literal["windows", "mac", "linux"], as_buffer: bool):
    pak = HGPAKFile(op.join(DATA_DIR, f"NMSARC.MeshPlanetSKY.{platform}.pak"), platform)
    fname = f"models/planets/sky/skysphere.geometry.mbin.{plat_map[platform]}"
    data = pak.extract_specific(fname, as_buffer)
    if as_buffer:
        assert isinstance(data, BytesIO)
        assert len(data.getvalue()) > 0
    else:
        assert isinstance(data, bytes)
        assert len(data) > 0


@pytest.mark.parametrize("platform", ("windows", "mac", "linux"))
@pytest.mark.parametrize("as_buffer", (True, False))
def test_extract_specific_multi_file(platform: Literal["windows", "mac", "linux"], as_buffer: bool):
    pak = HGPAKFile(op.join(DATA_DIR, f"NMSARC.MeshPlanetSKY.{platform}.pak"), platform)
    fnames = [
        f"models/planets/sky/skysphere.geometry.mbin.{plat_map[platform]}",
        f"models/planets/sky/rainbowplane.geometry.mbin.{plat_map[platform]}",
    ]
    data = pak.extract_specific(fnames, as_buffer)
    assert isinstance(data, dict)
    assert set(data.keys()) == set(fnames)
    if as_buffer:
        for blob in data.values():
            assert isinstance(blob, BytesIO)
            assert len(blob.getvalue()) > 0
    else:
        for blob in data.values():
            assert isinstance(blob, bytes)
            assert len(blob) > 0


def test_invalid_pak():
    with pytest.raises(InvalidFileException):
        with HGPAKFile(op.join(DATA_DIR, "NMSARC.MeshPlanetSKY.invalid.pak")):
            pass


# TODO: Add test for switch pak's.



import argparse
import importlib.util
import json
import logging
import os
import os.path as op
import pathlib
import platform
import sys
import time
from typing import Literal

from hgpaktool import __version__
from hgpaktool.api import HGPAKFile, InvalidFileException
from hgpaktool.constants import Platform, platform_map
from hgpaktool.utils import should_unpack

# Try import both lz4 and zstd.
# zstd is only required for windows, and lz4 is only required for mac, so if either fails, don't break
# immediately, only break once we know what platform we are targeting.

logger = logging.getLogger("hgpaktool")
logger.addHandler(logging.StreamHandler())

zstd_imported = False
try:
    if importlib.util.find_spec("zstandard") is not None:
        zstd_imported = True
except ModuleNotFoundError:
    pass

lz4_imported = False
try:
    if importlib.util.find_spec("lz4", "block") is not None:
        lz4_imported = True
except ModuleNotFoundError:
    pass


class SmartFormatter(argparse.HelpFormatter):
    # "Smaerter" help formatter c/o https://stackoverflow.com/a/22157136
    def _split_lines(self, text, width):
        if text.startswith("R|"):
            return text[2:].splitlines()
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)


class HGPAKNamespace(argparse.Namespace):
    plain: bool
    manifest: bool
    platform: Literal["windows", "mac", "switch", "linux"]
    compress: bool
    output: os.PathLike[str]
    filter: list[str]
    json: os.PathLike[str]
    dryrun: bool
    upper: bool
    unpack: bool
    repack: os.PathLike[str]
    filenames: list[str]
    list: bool
    hash: os.PathLike[str]
    no_hash_guid: bool
    quiet: bool
    verbose: int


def update_hashes(
    pak: HGPAKFile,
    hash_data: dict,
    pakname: str,
    args: HGPAKNamespace,
):
    pak_hash_data = {}
    for ifname, _hash in pak.get_hashes(args.filter, args.no_hash_guid):
        if args.upper:
            pak_hash_data[ifname.upper()] = _hash
        else:
            pak_hash_data[ifname] = _hash
    if pak_hash_data:
        if args.plain:
            hash_data.update(pak_hash_data)
        else:
            hash_data[op.basename(pakname)] = pak_hash_data


def run():
    parser = argparse.ArgumentParser(
        prog=f"HGPAKtool ({__version__})",
        description="A tool for handling HG's custom .pak format",
        formatter_class=SmartFormatter,
    )
    parser.add_argument(
        "-L",
        "--list",
        action="store_true",
        default=False,
        help="Generate a list of files contained within the pak file.",
    )
    parser.add_argument(
        "-p",
        "--plain",
        action="store_true",
        default=False,
        help=("Whether to output any generation informational files in a simplified format"),
    )
    parser.add_argument(
        "-M",
        "--manifest",
        action="store_true",
        default=False,
        help="Write the .manifest file when unpacking the .pak file. This is required for recompression.",
    )
    parser.add_argument(
        "--platform",
        choices=("windows", "mac", "linux", "switch"),
        default=platform_map[platform.system()],
        const=platform_map[platform.system()],
        nargs="?",
        help=(
            "R|The platform to unpack the files for. Default: %(default)s.\n"
            "Note: This changes the compression algorithm used to compress or decompress the files like so:\n"
            " - windows -> ZSTD\n"
            " - linux   -> ZSTD\n"
            " - macos   -> LZ4\n"
            " - switch  -> Oodle\n"
            "This will default to the platform you are on, so if you wish to decompress a file from a "
            "different platform or with a different compression algorithm, choose from the above list."
        ),
    )
    parser.add_argument(
        "-Z", "--compress", action="store_true", help="Whether or not to compress the provided files."
    )
    parser.add_argument(
        "-O",
        "--output",
        required=False,
        help=(
            "The directory to place extracted files in. If not provided, falls back to a folder called "
            "'EXTRACTED' in the current directory."
        ),
        type=pathlib.Path,
    )
    parser.add_argument(
        "-f",
        "--filter",
        action="append",
        help=(
            "R|A glob pattern which can be used to filter out the files which are to be extracted.\n"
            "This argument can be provided multiple times and the filters will be individually be applied to "
            "full set of files in each pak (ie. filters are OR'd, not AND'd)."
        ),
    )
    parser.add_argument(
        "-j",
        "--json",
        help=(
            "The path to a json file which can be used to indicate the files to be unpacked.\n"
            "The keys to the json are the paths (either relative or absolute) to the pak's that are to be "
            "extracted, and the values are the files within these pak's to extract.\n"
            "If the pak paths are relative, then a 'filename' argument MUST be passed which is the root "
            "directory of the provided pak names."
        ),
        type=pathlib.Path,
    )
    parser.add_argument(
        "--dryrun",
        action="store_true",
        default=False,
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--upper",
        action="store_true",
        default=False,
        help="If provided, extracted filenames will be converted to UPPERCASE.",
    )
    parser.add_argument(
        "--hash",
        help="The name of a file to dump the hash details of the contained files",
        type=pathlib.Path,
        metavar="OUTPUT.JSON",
    )
    parser.add_argument(
        "--no_hash_guid",
        action="store_true",
        default=False,
        help="Whether or not to hash the GUID for mbin files",
    )

    logging_group = parser.add_mutually_exclusive_group()
    logging_group.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase logging verbosity",
    )
    logging_group.add_argument("-q", "--quiet", action="store_true", help="Only log warnings and errors")

    pup_group = parser.add_mutually_exclusive_group()  # pup = pack/unpack
    pup_group.add_argument(
        "-U",
        "--unpack",
        action="store_true",
        help="Unpack the files from the provided pak files.",
    )
    pup_group.add_argument(
        "-R",
        "--repack",
        action="store_true",
        help="Repackthe specified manifest file.",
    )

    parser.add_argument(
        "filenames",
        nargs="+",
        help=(
            "The file(s) to pack or unpack. If this is a list of pak files or a directory, then it will be "
            "assumed that the files need to be unpacked.\nIf the filename is a single json file in the same "
            "format as produced by the -L flag, then it will also be unpacked as per the listed pak files "
            "and listed contents."
        ),
    )

    args = HGPAKNamespace()
    args = parser.parse_args(namespace=args)
    filenames = args.filenames

    generate_hashes = args.hash is not None

    # Handle logging level based on verbosity and quiet flags
    logger.setLevel(logging.INFO)
    if args.quiet:
        logger.setLevel(logging.ERROR)
    else:
        verbosity = args.verbose
        if verbosity >= 1:
            logger.setLevel(logging.DEBUG)

    if args.repack:
        mode = "repack"
    elif args.unpack:
        mode = "unpack"
    else:
        if generate_hashes:
            mode = "hash"
        elif should_unpack(filenames):
            # All the files provided are pak files, so decompress them unless we
            # have been asked to repack them
            mode = "unpack"
        else:
            mode = "repack"

    if args.platform == Platform.WINDOWS or args.platform == Platform.LINUX:
        if zstd_imported is False:
            logger.error(
                "You need to install zstandard for this code to work. Please run `pip install zstandard`"
            )
            sys.exit(1)
    elif args.platform == Platform.MAC:
        if lz4_imported is False:
            logger.error("You need to install lz4 for this code to work. Please run `pip install lz4`")
            sys.exit(1)
    elif args.platform == Platform.SWITCH:
        # Currently no need to do check anything for switch. Can probably do a check for oodle at some point.
        pass

    t1 = time.perf_counter()

    if mode == "unpack":
        output = op.abspath(args.output or "EXTRACTED")
        if not op.exists(output) and not args.list:
            os.makedirs(output, exist_ok=True)
        pack_count = 0
        file_count = 0
        filename_data: dict[str, list[str]] = {}

        json_file = None
        if args.json:
            json_file = args.json
        elif len(filenames) == 1 and filenames[0].lower().endswith(".json"):
            json_file = filenames[0]

        if generate_hashes:
            hash_data = {}

        if json_file is not None:
            root_dir = None
            if len(filenames) > 1:
                logger.error(
                    "Cannot unpack with json from multiple directories. Please only provide one directory"
                )
                sys.exit(1)
            elif len(filenames) == 1:
                if op.exists(filenames[0]) and op.isdir(filenames[0]):
                    root_dir = filenames[0]
            # In this case, we got a json file which contains a list of paks to unpack from.
            with open(json_file, "r") as f:
                json_data = json.load(f)
            for pak_path, req_contents in json_data.items():
                abs_pak_path = pak_path
                if not op.isabs(pak_path):
                    if root_dir is not None:
                        abs_pak_path = op.join(root_dir, pak_path)
                    else:
                        logger.error(
                            f"Cannot extract {pak_path} as it's a relative path. "
                            "Either provide the absolute path, or provide the root directory as the "
                            "'filename' argument."
                        )
                        continue
                if not abs_pak_path.lower().endswith(".pak"):
                    logger.debug(f"Skipping {pak_path}: Not a valid file to extract.")
                    continue
                try:
                    logger.debug(f"Reading {op.basename(abs_pak_path)}")
                    with HGPAKFile(abs_pak_path, args.platform) as pak:
                        file_count += pak.unpack(
                            output,
                            req_contents,
                            args.upper,
                            write_manifest=args.manifest,
                        )
                        if generate_hashes:
                            update_hashes(pak, hash_data, abs_pak_path, args)
                        pack_count += 1
                except InvalidFileException:
                    logger.debug(f"{abs_pak_path} is not a valid .pak file. Skipping")
        else:
            for filename in filenames:
                if op.isdir(filename):
                    for fname in os.listdir(filename):
                        if not fname.lower().endswith(".pak"):
                            logger.debug(f"{fname} is not a valid path to extract.")
                            continue
                        try:
                            logger.debug(f"Reading {fname}")
                            with HGPAKFile(op.join(filename, fname), args.platform) as pak:
                                if not args.list:
                                    file_count += pak.unpack(
                                        output,
                                        args.filter,
                                        args.upper,
                                        write_manifest=args.manifest,
                                    )
                                else:
                                    # Generate a list of the contained files
                                    fullpath = op.join(op.realpath(filename), fname)
                                    fnames = list(pak._get_filtered_filelist(args.filter).keys())
                                    if fnames:
                                        filename_data[fullpath] = fnames
                                if generate_hashes:
                                    update_hashes(pak, hash_data, fname, args)
                                pack_count += 1
                        except InvalidFileException:
                            logger.debug(f"{op.join(filename, fname)} is not a valid .pak file. Skipping")
                else:
                    try:
                        logger.debug(f"Reading {filename}")
                        with HGPAKFile(filename, args.platform) as pak:
                            if not args.list:
                                file_count += pak.unpack(
                                    output,
                                    args.filter,
                                    args.upper,
                                    write_manifest=args.manifest,
                                )
                            else:
                                # Generate a list of the contained files
                                fnames = list(pak._get_filtered_filelist(args.filter).keys())
                                if fnames:
                                    filename_data[op.realpath(filename)] = fnames
                            if generate_hashes:
                                update_hashes(pak, hash_data, filename, args)
                            pack_count += 1
                    except InvalidFileException:
                        logger.debug(f"{filename} is not a valid .pak file. Skipping")

        if args.list:
            if args.plain:
                with open("filenames.txt", "w") as f:
                    for pakname, filenames in filename_data.items():
                        f.write(f"Listing {pakname}\n")
                        for fname in filenames:
                            if args.upper:
                                f.write(fname.upper() + "\n")
                            else:
                                f.write(fname + "\n")
            else:
                with open("filenames.json", "w") as f:
                    if args.upper:
                        # Recase all the internal filenames
                        data = {}
                        for pakname, filenames in filename_data.items():
                            data[pakname] = [x.upper() for x in filenames]
                        f.write(json.dumps(data, indent=2))
                    else:
                        f.write(json.dumps(filename_data, indent=2))
            logger.info(f"Listed contents of {pack_count} .pak's in {time.perf_counter() - t1:.3f}s")
        else:
            logger.info(
                f"Unpacked {file_count} files from {pack_count} .pak's in {time.perf_counter() - t1:.3f}s"
            )
    elif mode == "hash":
        hash_data = {}
        pak_count = 0
        for filename in filenames:
            if op.isdir(filename):
                for fname in os.listdir(filename):
                    if not fname.lower().endswith(".pak"):
                        continue
                    try:
                        logger.debug(f"Reading {fname}")
                        with HGPAKFile(op.join(filename, fname), args.platform) as pak:
                            update_hashes(pak, hash_data, fname, args)
                            pak_count += 1
                    except InvalidFileException:
                        logger.debug(f"{op.join(filename, fname)} is not a valid .pak file. Skipping")
            else:
                try:
                    logger.debug(f"Reading {filename}")
                    with HGPAKFile(filename, args.platform) as pak:
                        update_hashes(pak, hash_data, filename, args)
                        pak_count += 1
                except InvalidFileException:
                    logger.debug(f"{op.join(filename, fname)} is not a valid .pak file. Skipping")
        if hash_data:
            with open(args.hash, "w") as f:
                json.dump(hash_data, f, indent=1)
        logger.info(f"Hashed contents of {pak_count} .pak's in {time.perf_counter() - t1:.3f}s")
    elif mode == "repack":
        if args.output:
            output = op.abspath(args.output)
            if op.isfile(output):
                logger.error(
                    "Provided output path is not a directory. Please provide a directory to repack files to."
                )
                sys.exit(1)
        else:
            output = None
        for i, fname in enumerate(filenames):
            HGPAKFile.repack(fname, output, args.compress, args.platform)
        logger.info(f"Repacked {i + 1} .pak's in {time.perf_counter() - t1:.3f}s")


if __name__ == "__main__":
    run()



import hashlib
import os
import os.path as op
import pathlib
from typing import Union


def determine_bins(num_bytes: int, bin_size: int = 0x10) -> int:
    """Determine the number of bins required to hold the requested number of bytes."""
    return (num_bytes + bin_size - 1) // bin_size


def reqChunkBytes(chunk_size: int) -> int:
    """Get the minimum required number of bytes which is a multiple of 0x10
    which can fit the specified number of bytes in the chunk."""
    return 0x10 * determine_bins(chunk_size)


def roundup(x: int) -> int:
    """Round up a number to the nearest 0x10 byte boundary."""
    # x >> 4 << 4 will round down to the nearest 0x10.
    # Then check if the number was higher by bitwise and-ing 0xF.
    # If it is, then add the 0x10.
    return (x >> 4 << 4) + ((x & 0xF) and 0x10)


def padding(x: int) -> int:
    """Determine the number of bytes required to pad the value to a 0x10
    byte boundary"""
    return (0x10 - (x & 0xF)) & 0xF


def should_unpack(filenames: list[str]) -> bool:
    """Determine whether we should unpack or not.
    This will return true if every file in the list has the extension .pak or we get a folder.
    """
    return (
        all([x.lower().endswith(".pak") for x in filenames])
        or (len(filenames) == 1 and op.isdir(filenames[0]))
        or (len(filenames) == 1 and filenames[0].lower().endswith(".json"))
    )


def hash_path(path: str) -> bytes:
    # This is the hash of the filename as per how the game generates them.
    return hashlib.md5(normalise_path(path).encode()).digest()


def normalise_path(path: str) -> str:
    return pathlib.PureWindowsPath(path).as_posix().lower()


def parse_manifest(manifest: Union[str, os.PathLike[str]]) -> list[str]:
    """Parse the manifest file and extract the list of files contained."""
    fnames = []
    with open(manifest, "r") as f:
        for line in f:
            sline = line.strip()
            if not sline:
                continue
            fnames.append(normalise_path(sline))
    return fnames



import os.path as op
import shutil
from pathlib import Path
from typing import Literal

import pytest
from utils import get_files

from hgpaktool import HGPAKFile

DATA_DIR = op.join(op.dirname(__file__), "data")


@pytest.mark.parametrize("fname,filecount", [("globals", 39), ("MeshPlanetSKY", 6)])
@pytest.mark.parametrize("platform", ("windows", "mac"))
def test_repack(tmp_path: Path, fname: str, filecount: int, platform: Literal["windows", "mac"]):
    # Ensure that repacked files are the same as the original

    # First, move the file to the temporary directory so that we can rename the file without the platform.
    shutil.copyfile(op.join(DATA_DIR, f"NMSARC.{fname}.{platform}.pak"), tmp_path / f"NMSARC.{fname}.pak")
    with HGPAKFile(tmp_path / f"NMSARC.{fname}.pak", platform) as pak:
        assert len(pak.filenames) == filecount
        # Extract the files to a temporary directory and analyse it.
        pak.unpack(tmp_path, write_manifest=True)

    files = get_files(tmp_path)
    assert len(files) == filecount + 1 + 1  # n files in pak. +1 for manifest and +1 for the original pak.

    # Rename the original pak file so that we can do a byte compare.
    shutil.move(tmp_path / f"NMSARC.{fname}.pak", op.join(tmp_path, f"NMSARC.{fname}.pak.original"))

    assert not op.exists(tmp_path / f"NMSARC.{fname}.pak")

    # Find the manifest file.
    manifest_fpath = None
    for fpath in files:
        if fpath.endswith(".manifest"):
            manifest_fpath = fpath
    assert manifest_fpath is not None
    assert op.exists(manifest_fpath)

    HGPAKFile.repack(manifest_fpath, platform=platform)
    assert op.exists(tmp_path / f"NMSARC.{fname}.pak")
    with open(tmp_path / f"NMSARC.{fname}.pak", "rb") as f:
        repacked_data = f.read()

    with open(tmp_path / f"NMSARC.{fname}.pak.original", "rb") as f:
        original_data = f.read()

    assert repacked_data == original_data



from io import BufferedWriter, BytesIO
from typing import Iterator

from hgpaktool.compressors import Compressor
from hgpaktool.utils import padding


def chunked_file_reader(fpaths: list[str], decompressed_chunk_size: int) -> Iterator[bytes]:
    """Yield chunks of size up to decompressed_chunk_size bytes from a file."""
    for fpath in fpaths:
        with open(fpath, "rb") as f:
            while True:
                data = f.read(decompressed_chunk_size)
                if not data:
                    break
                data_len = len(data)
                if data_len != decompressed_chunk_size:
                    # Add on the padding bytes
                    data += b"\x00" * padding(data_len)
                yield data


class FixedBuffer(BytesIO):
    def __init__(
        self,
        main_buffer: BufferedWriter,
        compressor: Compressor,
        compress: bool = False,
    ):
        self._clean_bytes = b"\x00" * compressor.decompressed_chunk_size
        super().__init__(self._clean_bytes)
        # The number of bytes remaining until we have a full buffer
        self.main_buffer = main_buffer
        self.compress = compress
        self.compressor = compressor
        self.remaining_bytes = self.compressor.decompressed_chunk_size
        # If we are compressing the data, keep track of the sizes of the
        # compressed data so we can write it into the TOC once we are done.
        self.compressed_block_sizes = []

    @property
    def _decompressed_chunk_size(self) -> int:
        return self.compressor.decompressed_chunk_size

    def add_bytes(self, data: bytes):
        """Add the provided bytes to the buffer.
        The amount of bytes passed in will never be more than 0x20000.
        If the amount fills the buffer then flush and prefill the next buffer.
        """
        data_size = len(data)
        written_bytes = self.write(data[: self.remaining_bytes])
        # Subtract of either the amount of bytes added or the number of
        # remaining bytes.
        self.remaining_bytes -= written_bytes
        if self.remaining_bytes == 0:
            self.write_to_main_buffer()
            # Reset the number of remaining bytes
            self.remaining_bytes = self.compressor.decompressed_chunk_size
        # If we have any extra bytes to write, write them now as the buffer is
        # currently empty.
        if data_size > written_bytes:
            new_written_bytes = self.write(data[written_bytes:])
            self.remaining_bytes -= new_written_bytes

    def write_to_main_buffer(self):
        """Write the data in the current buffer into the main buffer."""
        buffer = self.getbuffer()
        if self.compress:
            compressed_bytes = self.compressor.compress(buffer)
            compressed_size = len(compressed_bytes)
            if compressed_size >= self._decompressed_chunk_size:
                # If compression has somehow made it worse, use the original
                # bytes.
                self.main_buffer.write(self.getbuffer())
                compressed_size = self._decompressed_chunk_size
            else:
                self.main_buffer.write(compressed_bytes)
                # In this case we'll also need to write some extra bytes which
                # will make the next block we written at an address which is a
                # multiple of 0x10.
                self.main_buffer.write(b"\x00" * padding(compressed_size))
            self.compressed_block_sizes.append(compressed_size)
        else:
            self.main_buffer.write(buffer)
        buffer.release()
        # Once we have written this to the parent buffer, clear ourselves to be
        # ready for the next lot of bytes.
        self.clear()

    def clear(self):
        """Clear the current buffer."""
        self.seek(0)
        self.write(self._clean_bytes)
        self.seek(0)



import fnmatch
import hashlib
import os
import os.path as op
import struct
from collections import namedtuple
from dataclasses import dataclass
from functools import lru_cache
from io import SEEK_SET, BufferedReader, BufferedWriter, BytesIO
from logging import NullHandler, getLogger
from typing import Iterable, Literal, Mapping, Optional, Union, overload

from hgpaktool.buffers import FixedBuffer, chunked_file_reader
from hgpaktool.compressors import Compressor
from hgpaktool.constants import (
    Compression,
    Platform,
    PlatformLiteral,
    compression_map,
)
from hgpaktool.utils import (
    determine_bins,
    hash_path,
    normalise_path,
    padding,
    parse_manifest,
    reqChunkBytes,
    roundup,
)


@dataclass
class FileInfo:
    file_hash_: bytes
    start_offset: int
    decompressed_size: int

    @property
    def file_hash(self) -> int:
        return int.from_bytes(self.file_hash_, "big", signed=False)

    def values(self):
        return (self.file_hash_, self.start_offset, self.decompressed_size)

    def __str__(self):
        return (
            f"Hash: 0x{self.file_hash:X}, offset: 0x{self.start_offset:X}, "
            f"decompressed size: 0x{self.decompressed_size:X}"
        )


FILEINFO_FMT = "<16s2Q"
CHUNKINFO = namedtuple("CHUNKINFO", ["size", "offset"])

HGPAKFMT_VERSION = 2


logger = getLogger(__name__)
logger.addHandler(NullHandler())


class InvalidFileException(Exception):
    pass


class PackedFile:
    """This represents a packed file within the HGPAK file"""

    __slots__ = ("offset", "size", "path", "_in_chunks", "_decompressed_chunk_size")

    def __init__(self, offset: int, size: int, path: str, decompressed_chunk_size: int):
        self.offset = offset
        self.size = size
        self.path = path
        self._decompressed_chunk_size = decompressed_chunk_size
        self._in_chunks = None

    @property
    def filename_hash(self):
        # This hash is the md5 hash of the name.
        return hash_path(self.path)

    @property
    def in_chunks(self) -> tuple[int, int]:
        """Determine which chunks the file is contained in."""
        if self._in_chunks is None:
            if self.offset % self._decompressed_chunk_size == 0:
                start_chunk = determine_bins(self.offset, self._decompressed_chunk_size)
            else:
                start_chunk = determine_bins(self.offset, self._decompressed_chunk_size) - 1
            end_chunk = determine_bins(self.offset + self.size, self._decompressed_chunk_size) - 1
            self._in_chunks = (start_chunk, end_chunk)
        return self._in_chunks

    @property
    def first_chunk_offset(self):
        """The offset within the first chunk that the file starts at."""
        return self.offset % self._decompressed_chunk_size

    @property
    def last_chunk_offset_end(self):
        """The offset within the last chunk where the file ends."""
        return (self.offset + self.size) % self._decompressed_chunk_size

    def __str__(self):
        return (
            f"File: {self.path}: Offset: 0x{self.offset:X}, Size: 0x{self.size:X}, "
            f"In chunks: {self.in_chunks}"
        )

    def __repr__(self):
        return str(self)


class HGPakHeader:
    def __init__(self):
        self.version = HGPAKFMT_VERSION
        self.file_count = 0
        self.chunk_count = 0
        self.is_compressed = False
        self.data_offset = 0

    def read(self, fobj: BufferedReader):
        fobj.seek(0, SEEK_SET)
        if struct.unpack("5s", fobj.read(5))[0] != b"HGPAK":
            raise InvalidFileException(f"{fobj.name} does not appear to be a valid HGPAK file.")
        fobj.seek(8, SEEK_SET)
        self.version, self.file_count, self.chunk_count, self.is_compressed, self.data_offset = struct.unpack(
            "<QQQ?7xQ", fobj.read(0x28)
        )
        if self.version != HGPAKFMT_VERSION:
            raise InvalidFileException(f"{fobj.name} does not appear to be a valid HGPAK file.")

    def write(self, fobj: BufferedWriter):
        # First, write the magic
        fobj.write(struct.pack("<5s3x", b"HGPAK"))
        fobj.write(
            struct.pack(
                "<QQQ?7xQ",
                self.version,
                self.file_count,
                self.chunk_count,
                self.is_compressed,
                self.data_offset,
            )
        )

    def __str__(self):
        return (
            f"HGPak Header:\n"
            f" Version {self.version}\n"
            f" Files: {self.file_count}\n"
            f" Chunks: {self.chunk_count}\n"
            f" is Compressed: {self.is_compressed}\n"
            f" Data offset: 0x{self.data_offset:X}\n"
        )


class HGPakFileIndex:
    def __init__(self):
        self.fileInfo: list[FileInfo] = []

    def read(self, file_count: int, fobj: BufferedReader, n: int = -1):
        """Read up to n entries from the file index. If n is -1 it will read all."""
        for i in range(file_count):
            if n != -1 and i >= n:
                return
            finf = FileInfo(*struct.unpack(FILEINFO_FMT, fobj.read(0x20)))
            self.fileInfo.append(finf)

    def write(self, fobj: BufferedWriter):
        for finf in self.fileInfo:
            fobj.write(struct.pack(FILEINFO_FMT, *finf.values()))

    def __str__(self):
        res = "File Index\n----------\n"
        res += "\n".join([str(finf) for finf in self.fileInfo])
        return res


class HGPakChunkIndex:
    chunk_sizes: tuple[int]

    def __init__(self):
        self.chunk_offset: list[int] = []

    def read(self, chunk_count: int, fobj: BufferedReader):
        self.chunk_sizes = struct.unpack(f"<{chunk_count}Q", fobj.read(8 * chunk_count))

    def calculate_chunk_offsets(self, start: int):
        # Calculate the chunk offsets based on sizes and initial start location.
        curr_pos = start
        for size in self.chunk_sizes:
            self.chunk_offset.append(curr_pos)
            curr_pos += reqChunkBytes(size)


class HGPAKFile:
    fobj: BufferedReader

    def __init__(
        self,
        filepath: Union[str, os.PathLike[str]],
        platform: Union[Platform, PlatformLiteral] = Platform.WINDOWS,
    ):
        self.compressor = Compressor(compression_map.get(platform, Compression.ZSTD))
        self.fpath = filepath

        self.index: list[str] = []  # The list of files contained
        self.header = HGPakHeader()
        self.fileIndex = HGPakFileIndex()
        self.chunkIndex = HGPakChunkIndex()
        self.files: dict[str, PackedFile] = {}
        self.filenames: list[str] = []

    def __enter__(self):
        # Open the provided path and build the index.
        self.fobj = open(self.fpath, "rb")
        self._parse()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.fobj is not None:
            self.fobj.close()

    @property
    def _decompressed_chunk_size(self) -> int:
        return self.compressor.decompressed_chunk_size

    @property
    def pak_name(self) -> str:
        return op.basename(self.fpath)

    @property
    def _extractor_function(self):
        if self.header.is_compressed:
            return self._extract_file_compressed
        else:
            return self._extract_file_uncompressed

    def dump_index(self, dest: Union[str, os.PathLike[str]]):
        """Dump the fileindex data of the pak file to disk

        Parameters
        ----------
        dest
            The file to dump the info into
        """
        with open(dest, "w") as f:
            for filename in self.filenames:
                pf = self.files[filename]
                f.write(f"{pf}\n")

    def _parse(self):
        if self.fobj is None:
            raise Exception("HPAKFile has no initialised fobj")
        self.header.read(self.fobj)
        self.fileIndex.read(self.header.file_count, self.fobj)
        if self.header.is_compressed is False:
            # We only need to read the filename data and then return.
            self.fobj.seek(self.header.data_offset, SEEK_SET)
            filename_data = self.fobj.read(self.fileIndex.fileInfo[0].decompressed_size)
            self.filenames = [
                x.decode()
                for x in filename_data[: self.fileIndex.fileInfo[0].decompressed_size]
                .rstrip(b"\r\n")
                .split(b"\r\n")
            ]
            for i, fname in enumerate(self.filenames):
                if fname:
                    finf = self.fileIndex.fileInfo[i + 1]
                    self.files[fname] = PackedFile(
                        finf.start_offset,
                        finf.decompressed_size,
                        fname,
                        self._decompressed_chunk_size,
                    )
            return

        if self.header.is_compressed:
            self.chunkIndex.read(self.header.chunk_count, self.fobj)
        self.chunkIndex.calculate_chunk_offsets(self.header.data_offset)

        # Determine how many chunks to decompress to read the filenames.
        chunks_for_filenames = determine_bins(
            self.fileIndex.fileInfo[0].decompressed_size, self._decompressed_chunk_size
        )

        # Decompress these chunks to read the filenames.
        manifest = b""
        for i in range(chunks_for_filenames):
            if (_chunk := self._decompress_chunk(i)) is not None:
                manifest += _chunk
            else:
                logger.error(f"There was an error reading the filename section for {self.fpath}")
        self.filenames = [
            x.decode()
            for x in manifest[: self.fileIndex.fileInfo[0].decompressed_size].rstrip(b"\r\n").split(b"\r\n")
        ]
        assert len(self.filenames) == self.header.file_count - 1, "file count mismatch"
        for i, fname in enumerate(self.filenames):
            if fname:
                finf = self.fileIndex.fileInfo[i + 1]
                self.files[fname] = PackedFile(
                    finf.start_offset - self.header.data_offset,
                    finf.decompressed_size,
                    fname,
                    self._decompressed_chunk_size,
                )

    @lru_cache(maxsize=256)
    def _decompress_chunk(self, chunkIdx: int) -> Optional[bytes]:
        self.fobj.seek(self.chunkIndex.chunk_offset[chunkIdx], SEEK_SET)
        chunk_size = self.chunkIndex.chunk_sizes[chunkIdx]
        return self.compressor.decompress(self.fobj.read(chunk_size))

    def _get_filtered_filelist(
        self, filters: Union[list[str], str, None] = None
    ) -> Mapping[str, Optional[PackedFile]]:
        """Filter the known file list.
        This returns a dictionary instead of a set so that the order is always the same.
        """
        files = {}
        if filters is not None:
            if isinstance(filters, str):
                if "*" in filters:
                    for filtered in fnmatch.filter(self.files, filters.lower()):
                        files[filtered] = None
                else:
                    files[filters.lower()] = None
            else:
                for filter_ in filters:
                    if "*" in filter_:
                        for filtered in fnmatch.filter(self.files, filter_.lower()):
                            files[filtered] = None
                    else:
                        files[filter_.lower()] = None
        else:
            files = self.files
        return files

    def get_hashes(
        self,
        filters: Union[list[str], str, None] = None,
        mask_guid: bool = False,
        algorithm: str = "md5",
    ) -> Iterable[tuple[str, str]]:
        """Generate hashes for the specified file(s) in the pak.

        Parameters
        ----------
        filters:
            An optional list of glob patterns to pattern match against when extracting.
            Only files which match the pattern will be extracted.
            This can also just be a single string in which case just this file will be extracted.
        mask_guid:
            If True, the GUID in the mbin file will be masked.
            This will only happen if the file is an mbin file (checked base on the magic), otherwise it will
            do nothing.
        algorithm:
            The name of the algorithm. This must be one provided by python (cf. hashlib.algorithms_available).

        Returns
        -------
        An iterable over the hashes.
        This will always be an iterable even if a single filename is passed in.
        """
        files = self._get_filtered_filelist(filters)

        if len(files) == 0:
            return

        func = self._extractor_function
        _base_hash = hashlib.new(algorithm)
        for fpath in files:
            _hash = _base_hash.copy()
            for i, chunk in enumerate(func(fpath)):
                if len(chunk) == 0:
                    continue
                if i == 0:
                    if mask_guid:
                        magic = chunk[:8]
                        if (
                            magic == b"\xcc\xcc\xcc\xcc\xcc\xcc\xcc\xcc"
                            or magic == b"\xdd\xdd\xdd\xdd\xdd\xdd\xdd\xdd"
                        ):
                            chunk = chunk[:0x10] + b"\x00" * 8 + chunk[0x18:]
                _hash.update(chunk)
            yield (fpath, _hash.hexdigest().upper())

    @overload
    def extract_specific(self, filepaths: list[str], as_buffer: Literal[True]) -> Mapping[str, BytesIO]: ...

    @overload
    def extract_specific(self, filepaths: list[str], as_buffer: Literal[False]) -> Mapping[str, bytes]: ...

    @overload
    def extract_specific(self, filepaths: str, as_buffer: Literal[True]) -> BytesIO: ...

    @overload
    def extract_specific(self, filepaths: str, as_buffer: Literal[False]) -> bytes: ...

    def extract_specific(
        self,
        filepaths: Union[list[str], str],
        as_buffer: bool = True,
    ) -> Union[Mapping[str, Union[bytes, BytesIO]], Union[bytes, BytesIO]]:
        """Extract a specific subset of the files in a pak file.
        This is optimised to only get the info required to extract these files, so any extra metadata that
        we'd normally be able to get from the instance will not be correct.
        Only use this when you have a fixed set of files to extract with concrete filepaths."""
        if isinstance(filepaths, list):
            fname_hash_map = {hash_path(filepath): filepath for filepath in filepaths}
        else:
            fname_hash_map = {hash_path(filepaths): filepaths}
        # Go through the process semi-manually, skipping anything we don't need.
        files_to_read = len(fname_hash_map)
        files_read = 0
        file_infos: dict[str, FileInfo] = {}
        out_data: dict[str, Union[bytes, BytesIO]] = {}

        try:
            self.fobj = open(self.fpath, "rb")
            self.header = HGPakHeader()
            self.header.read(self.fobj)
            # Read the file index, but only load data if it has the same hash.
            for _ in range(self.header.file_count):
                if files_read == files_to_read:
                    break
                _hash = struct.unpack_from("<16s", buffer=self.fobj.read(0x10))[0]
                if _hash in fname_hash_map:
                    fname = fname_hash_map[_hash]
                    finf = FileInfo(_hash, *struct.unpack("<2Q", self.fobj.read(0x10)))
                    file_infos[fname] = finf
                    files_read += 1
                else:
                    self.fobj.seek(0x10, 1)

            max_chunk_index = 0

            # Populate the `self.files` object with just the files we care about.
            for fname, finf in file_infos.items():
                pf = PackedFile(
                    finf.start_offset - self.header.data_offset,
                    finf.decompressed_size,
                    fname,
                    self._decompressed_chunk_size,
                )
                max_chunk_index = max(max_chunk_index, pf.in_chunks[1] + 1)
                self.files[fname] = pf
            if not max_chunk_index:
                max_chunk_index = self.header.chunk_count

            # Jump to the start of the chunk index
            self.fobj.seek(0x30 + self.header.file_count * 0x20, 0)
            if self.header.is_compressed:
                self.chunkIndex.read(max_chunk_index, self.fobj)
            self.chunkIndex.calculate_chunk_offsets(self.header.data_offset)

            # Extract the actual data.
            # Depending on whether we get just one filepath to extract or more, we'll either add to the dict
            # or return immediately.
            func = self._extractor_function
            for fpath in self.files:
                buffer = BytesIO()
                for chunk in func(fpath):
                    buffer.write(chunk)
                if buffer.tell() > 0:
                    if as_buffer:
                        if isinstance(filepaths, str):
                            return buffer
                        else:
                            out_data[fpath] = buffer
                    else:
                        if isinstance(filepaths, str):
                            return buffer.getvalue()
                        else:
                            out_data[fpath] = buffer.getvalue()
                else:
                    continue

            if not isinstance(filepaths, str):
                return out_data
            else:
                # If we get here, we never found the file so we'll return empty bytes.
                logger.error(f"Unable to find the file {filepaths!r} in {self.fpath}")
                return b""
        finally:
            self.fobj.close()

    def extract(
        self,
        filters: Union[list[str], str, None] = None,
        max_bytes: int = -1,
    ) -> Iterable[tuple[str, bytes]]:
        """Extract the specified file(s) out of the pak iteratively.

        Parameters
        ----------
        filters:
            An optional list of glob patterns to pattern match against when extracting.
            Only files which match the pattern will be extracted.
            This can also just be a single string in which case just this file will be extracted.
        max_bytes:
            Maximum number of bytes to extract per-file. If this is -1 (the default) it will extract the
            entire file.

        Returns
        -------
        An iterable over the extracted files.
        This will always be an iterable even if a single filename is passed in.
        """
        files = self._get_filtered_filelist(filters)

        if len(files) == 0:
            return

        func = self._extractor_function
        for fpath in files:
            buffer = BytesIO()
            for chunk in func(fpath, max_bytes):
                buffer.write(chunk)
            if buffer.tell() > 0:
                if max_bytes != -1:
                    yield (fpath, buffer.getvalue()[:max_bytes])
                else:
                    yield (fpath, buffer.getvalue())
            else:
                continue

    def unpack(
        self,
        dest: Union[str, os.PathLike[str]],
        filters: Union[list[str], str, None] = None,
        upper: bool = False,
        max_bytes: int = -1,
        write_manifest: bool = False,
    ) -> int:
        """Unpack the contained files to the specified destination

        Parameters
        ----------
        dest:
            The target folder to extract the files to.
        filters:
            An optional list of glob patterns to pattern match against when unpacking.
            Only files which match the pattern will be unpacked.
            This can also just be a single string in which case just this file will be unpacked.
        upper:
            If True, file names will be normalised to upper case.
            Note: This setting will be ignored if ``write_manifest`` is True.
        max_bytes:
            Maximum number of bytes to unpack per-file. If this is -1 (the default) it will extract the entire
            file.
        write_manifest:
            Whether or not to write the manifest for the pak to disk.
            This is required if you want to repack the archive.
            This file will be written at the top level of the extraction directory.

        Returns
        -------
        Total number of files unpacked.
        """
        files = self._get_filtered_filelist(filters)

        if len(files) == 0:
            return 0

        if upper and write_manifest:
            logger.warning(
                "`upper` and `write_manifest` arguments are both set to True. This combination is not valid."
                " The value for `upper` will be ignored."
            )

        # Loop over the files to extract their contained data.
        i = 0
        func = self._extractor_function
        for fpath in files:
            _export_path, fname = op.split(fpath)
            dir_ = op.join(dest, _export_path)
            # If we are writing the manifest, ignore the upper flag.
            if upper and not write_manifest:
                dir_ = op.join(dest, _export_path.upper())
                fname = fname.upper()
            if dir_:
                os.makedirs(dir_, exist_ok=True)
            # Open the file and then loop over returned chunks and write.
            # This is more efficient, especially for compressed data than decompressing the entire file and
            # then writing it.
            with open(op.join(dir_, fname), "wb") as f:
                for chunk in func(fpath, max_bytes):
                    f.write(chunk)

            i += 1

        if write_manifest:
            with open(op.join(dest, f"{self.pak_name}.manifest"), "wb") as f:
                for fname in files:
                    f.write(fname.encode() + b"\r\n")

        return i

    def _extract_file_compressed(self, fpath: str, max_bytes: int = -1) -> Iterable[bytes]:
        # Extract compressed chunks from the pak file.

        # First, get the file info.
        finf = self.files.get(fpath)
        if not finf:
            raise FileNotFoundError(f"The specified file path ({fpath!r}) doesn't exist in this pak")
        start_chunk, end_chunk = finf.in_chunks
        first_off = finf.first_chunk_offset
        last_off = finf.last_chunk_offset_end

        bytes_read = 0

        if start_chunk == end_chunk:
            # The data is contained entirely within the same chunk.
            decompressed = self._decompress_chunk(start_chunk)
            if decompressed is None:
                logger.error(f"There was an issue decompressing chunk {start_chunk}")
                logger.error(f"Unable to extract file: {fpath}")
                return
            if last_off:
                bytes_read += last_off - first_off
                yield decompressed[first_off:last_off]
            else:
                bytes_read += len(decompressed) - first_off
                yield decompressed[first_off:]
        else:
            for chunk_idx in range(start_chunk, end_chunk + 1):
                decompressed = self._decompress_chunk(chunk_idx)
                if decompressed is None:
                    logger.error(f"There was an issue decompressing chunk {chunk_idx}")
                    logger.error(f"Unable to extract file: {fpath}")
                    return
                if chunk_idx == start_chunk:
                    bytes_read += len(decompressed) - first_off
                    yield decompressed[first_off:]
                elif chunk_idx == end_chunk:
                    if not last_off:
                        bytes_read += len(decompressed)
                        yield decompressed
                    else:
                        bytes_read += last_off
                        yield decompressed[:last_off]
                else:
                    bytes_read += len(decompressed)
                    yield decompressed
                if max_bytes != -1:
                    # Once we have at least as many bytes as max_bytes, then stop.
                    if bytes_read >= max_bytes:
                        break

    def _extract_file_uncompressed(self, fpath: str, max_bytes: int = -1) -> Iterable[bytes]:
        # Extract uncompressed chunks from the pak file.
        finf = self.files.get(fpath)
        if not finf:
            raise FileNotFoundError(f"The specified file path ({fpath!r}) doesn't exist in this pak")
        self.fobj.seek(finf.offset)

        extract_size = finf.size
        if max_bytes != -1:
            extract_size = max_bytes

        # Read each of the chunks
        chunks = extract_size // self._decompressed_chunk_size
        if not chunks:
            yield self.fobj.read(extract_size)
        else:
            # First, read all the full chunk data.
            for _ in range(chunks):
                yield self.fobj.read(self._decompressed_chunk_size)
            # Then finally, read the remainder is there is any.
            if (rem := extract_size % self._decompressed_chunk_size) != 0:
                yield self.fobj.read(rem)

    def _pack_files(
        self,
        filepaths: list[str],
        root_dir: Union[str, os.PathLike[str]],
        pak_fpath: Union[str, os.PathLike[str]],
        compress: bool = True,
    ):
        """Pack the provided files.
        The first file MUST be a .manifest file containing a list of the files in the .pak file.
        Note that since this is an internal method this will not be validated."""
        fullpaths: list[str] = []

        # Loop over the filepaths and get their name hashes and sizes.
        for fpath in filepaths:
            realpath = op.realpath(op.join(root_dir, fpath))
            fullpaths.append(realpath)
            self.fileIndex.fileInfo.append(FileInfo(hash_path(fpath), 0, os.stat(realpath).st_size))
        file_count = len(self.fileIndex.fileInfo)

        # Determine the offsets of each file in uncompressed space.
        curr_total_data_size = 0
        for finfo in self.fileIndex.fileInfo:
            finfo.start_offset = curr_total_data_size
            curr_total_data_size += roundup(finfo.decompressed_size)

        # Now that we have the total uncompressed data, we may determine the total number of chunks.
        chunk_count = determine_bins(curr_total_data_size, self._decompressed_chunk_size)

        # Now that we know the total chunk count and file count, we know how big the "pre-data" data is.
        data_offset = 0x30 + 0x20 * file_count + compress * 0x8 * chunk_count
        # Also add padding so that the data always starts at an offset which is a
        # multiple of 0x10
        extra_padding = padding(data_offset)
        data_offset += extra_padding

        # Fixup the offsets now that we know the data offset.
        for finfo in self.fileIndex.fileInfo:
            finfo.start_offset += data_offset

        # Write the data back to the file.
        self.header.file_count = file_count
        self.header.chunk_count = chunk_count
        self.header.is_compressed = compress
        self.header.data_offset = data_offset

        with open(pak_fpath, "wb") as f:
            self.header.write(f)
            self.fileIndex.write(f)
            # Reserve space for the compressed chunk sizes if the file is compressed.
            chunk_index_offset = f.tell()
            if compress:
                f.write(b"\x00" * (8 * chunk_count + extra_padding))

            sub_buffer = FixedBuffer(f, self.compressor, compress)

            # Write all the files into the pak.
            for _data in chunked_file_reader(fullpaths, self._decompressed_chunk_size):
                sub_buffer.add_bytes(_data)
            # Finally, call write_to_main_buffer to flush the last block to the file.
            sub_buffer.write_to_main_buffer()

            if compress:
                f.seek(chunk_index_offset, SEEK_SET)
                # Get the compressed block sizes and write to the chunk info section.
                for chunk_size in sub_buffer.compressed_block_sizes:
                    f.write(struct.pack("<Q", chunk_size))

    @classmethod
    def repack(
        cls,
        manifest: Union[str, os.PathLike[str]],
        out_fpath: Optional[Union[str, os.PathLike[str]]] = None,
        compress: bool = True,
        platform: Union[Platform, PlatformLiteral] = Platform.WINDOWS,
    ):
        """Repack the provided manifest file

        Parameters
        ----------
        manifest:
            The path to the manifest file to be repacked.
            This file should be the original file which was unpacked by HGPAKTool.
            If this file is modified it MUST use CRLF line endings and have a trailing line ending.
        out_fpath:
            The destination pak path. If not provided this will fallback to the name of the pak based on the
            manifest and will be written to the same directory as the manifest.
        compress:
            Whether or not to compress the data contained within the pak file.
            Compression format will depend on what platform is provided.
        platform:
            The target platform for the pak file.
        """
        # Generate the required paths based on the manifest file. It's basically our "source of truth"
        real_manifest_path = op.realpath(manifest)
        manifest_dir = op.dirname(real_manifest_path)
        manifest_name = op.basename(manifest)
        pak_name, _ = op.splitext(manifest_name)
        pak = cls(pak_name, platform=platform)
        # Parse the manifest so that we have a file list:
        file_list = parse_manifest(manifest)
        if out_fpath is None:
            out_fpath = op.join(manifest_dir, pak_name)
        pak._pack_files(
            [normalise_path(op.basename(manifest)), *file_list],
            manifest_dir,
            out_fpath,
            compress,
        )



import os.path as op
import urllib.request
from platform import system as os_name

OSNAMEMAP = {"Windows": "WIN", "Linux": "LNX", "Darwin": "MAC"}


class OSConstMap:
    LIB_NAME: str
    LIB_URL: str

    """ Class which provides values of constants in an OS-dependent manner """

    def __init__(self):
        self.os = OSNAMEMAP.get(os_name(), "WIN")

    def __setattr__(self, name, value):
        self.__dict__[name] = value

    def __getattr__(self, name):
        """Retreive the required value

        Names will be suffixed by '_<OS>', so we want to just get the os
        dependent version
        If the variable is assigned in a non-OS-dependent manner (ie. no
        suffix then this will not be called and the value will be retreived
        directly)

        Parameters
        ----------
        name : str
            Name of the variable without an OS identifier
        """
        os_dep_name = name + "_" + self.os
        return self.__dict__[os_dep_name]


OSCONST = OSConstMap()

# Binaries urls
OSCONST.LIB_URL_WIN = "https://raw.githubusercontent.com/WorkingRobot/OodleUE/main/Engine/Source/Programs/Shared/EpicGames.Oodle/Sdk/2.9.3/win/redist/oo2core_9_win64.dll"
# OSCONST.LIB_URL_WIN = "https://origin.warframe.com/origin/E926E926/Tools/Oodle/x64/final/oo2core_9_win64.dll.F2DB01967705B62AECEF3CD3E5A28E4D.lzma"
OSCONST.LIB_URL_MAC = "https://raw.githubusercontent.com/WorkingRobot/OodleUE/main/Engine/Source/Runtime/OodleDataCompression/Sdks/2.9.8/lib/Mac/liboo2coremac64.2.9.8.dylib"

# Binaries names
OSCONST.LIB_NAME_WIN = "oo2core_9_win64.dll"
OSCONST.LIB_NAME_MAC = "liboo2coremac64.2.9.8.dylib"


def download_dll(out_dir: str = ""):
    """Simple script to download the Oodle binary."""
    with urllib.request.urlopen(OSCONST.LIB_URL) as f:
        print(f"Downloading Oodle from {OSCONST.LIB_URL}")
        data = f.read()
        # decompressed = lzma.decompress(f.read())
    out_path = op.join(out_dir, OSCONST.LIB_NAME)
    if op.exists(out_path):
        inp = input("This file already exists. Do you want to override? (Y/N): ")
        if inp.lower() != "y":
            return
    with open(out_path, "wb") as f:
        f.write(data)
    print(f"Wrote file to {op.join(out_dir, OSCONST.LIB_NAME)}")


if __name__ == "__main__":
    # For now, running this script as main will simply download the lib into
    # your cwd.
    download_dll()



from collections import defaultdict
from enum import Enum
from typing import Literal


class Platform(str, Enum):
    WINDOWS = "windows"
    MAC = "mac"
    LINUX = "linux"
    SWITCH = "switch"


PlatformLiteral = Literal["windows", "mac", "switch", "linux"]


platform_map = defaultdict(
    lambda: "windows",
    {
        "Windows": Platform.WINDOWS.value,
        "Linux": Platform.LINUX.value,
        "Darwin": Platform.MAC.value,
    },
)


class Compression(str, Enum):
    ZSTD = "zstd"
    LZ4 = "lz4"
    OODLE = "oodle"


CompressionLiteral = Literal["zstd", "lz4", "oodle"]


compression_map = defaultdict(
    lambda: Compression.ZSTD,
    {
        "windows": Compression.ZSTD,
        "linux": Compression.ZSTD,
        "mac": Compression.LZ4,
        "switch": Compression.OODLE,
    },
)



import os.path as op
import sys
from logging import NullHandler, getLogger
from typing import Optional, Union, cast

from hgpaktool.constants import Compression, CompressionLiteral
from hgpaktool.oodle import OodleCompressor, OodleDecompressionError
from hgpaktool.os_funcs import OSCONST

# Try import both lz4 and zstd.
# zstd is only required for windows, and lz4 is only required for mac, so if either fails, don't break
# immediately, since we may not need it. An error will already be raised by the api if attempting to read a
# file without the required libraries installed.
try:
    import zstandard as zstd
except ModuleNotFoundError:
    pass

try:
    import lz4.block
except ModuleNotFoundError:
    pass


logger = getLogger(__name__)
logger.addHandler(NullHandler())


class Compressor:
    def __init__(self, compression: Union[Compression, CompressionLiteral] = Compression.ZSTD):
        self.compression = compression
        if self.compression == Compression.ZSTD:
            # TEMP fix for decompression. Won't work for compression.
            self.compressor = zstd.ZstdDecompressor()
            self.decompressed_chunk_size = 0x10000
            self._decompress_func = self._decompress_windows
        elif self.compression == Compression.LZ4:
            self.compressor = lz4.block
            self.decompressed_chunk_size = 0x20000
            self._decompress_func = self._decompress_mac
        else:
            self.compressor = OodleCompressor(op.join(op.dirname(__file__), "lib", OSCONST.LIB_NAME))
            self.decompressed_chunk_size = 0x20000
            self._decompress_func = self._decompress_switch

    def compress(self, buffer: memoryview) -> bytes:
        if self.compression == Compression.ZSTD:
            self.compressor = zstd.ZstdCompressor(
                compression_params=zstd.ZstdCompressionParameters(
                    compression_level=14,
                    write_content_size=True,
                ),
            )
            return self.compressor.compress(
                buffer,
            )
        elif self.compression == Compression.LZ4:
            self.compressor = cast(lz4.block, self.compressor)
            return self.compressor.compress(
                buffer,
                mode="high_compression",
                store_size=False,
                compression=10,
            )
        else:
            self.compressor = cast(OodleCompressor, self.compressor)
            return self.compressor.compress(buffer.tobytes("A"), self.decompressed_chunk_size)

    def _decompress_windows(self, data: bytes) -> Optional[bytes]:
        self.compressor = cast(zstd.ZstdDecompressor, self.compressor)
        try:
            return self.compressor.decompress(data)
        except zstd.ZstdError:
            if len(data) == self.decompressed_chunk_size:
                # In this case the block was just not compressed. Return it.
                return data
            else:
                if data[:2] == b"\x8c\x0a":
                    logger.error("Provided pak is from the switch. Please add `--platform switch`")
                    sys.exit(1)
                logger.exception("Error decompressing a chunk:")
                return None

    def _decompress_mac(self, data: bytes) -> Optional[bytes]:
        self.compressor = cast(lz4.block, self.compressor)
        try:
            return self.compressor.decompress(data, uncompressed_size=self.decompressed_chunk_size)
        except lz4.block.LZ4BlockError:
            if len(data) == self.decompressed_chunk_size:
                # In this case the block was just not compressed. Return it.
                return data
            else:
                if data[:2] == b"\x8c\x0a":
                    logger.error("Provided pak is from the switch. Please add `--platform switch`")
                    sys.exit(1)
                logger.exception("Error decompressing a chunk:")
                return None

    def _decompress_switch(self, data: bytes) -> Optional[bytes]:
        self.compressor = cast(OodleCompressor, self.compressor)
        try:
            return self.compressor.decompress(data, len(data), self.decompressed_chunk_size)
        except OodleDecompressionError:
            if len(data) == self.decompressed_chunk_size:
                # In this case the block was just not compressed. Return it.
                return data
            else:
                logger.exception("Error decompressing a chunk:")
                return None

    def decompress(self, data: bytes) -> Optional[bytes]:
        """Decompress the provided data blob. This will be decompressed based on the platform specified for
        this compressor object."""
        return self._decompress_func(data)



import os.path as op
from ctypes import c_char_p, cdll, create_string_buffer

# TODO:
# - Add the compression levels as an enum
# - Add the compression algorithms as an enum
# https://gist.github.com/Nukem9/4ab163e7d38ae22c09be8a31586a6edf


class OodleDecompressionError(Exception):
    pass


class OodleCompressionError(Exception):
    pass


class OodleCompressor:
    """
    Oodle decompression implementation.
    Requires Windows and the external Oodle library.
    """

    def __init__(self, library_path: str) -> None:
        """
        Initialize instance and try to load the library.
        """
        if not op.exists(library_path):
            raise Exception("Could not open Oodle DLL, make sure it is configured correctly.")

        try:
            self.handle = cdll.LoadLibrary(library_path)
        except OSError as e:
            raise Exception("Could not load Oodle DLL, requires Windows and 64bit python to run.") from e

    def compress(self, payload: bytes, size: int) -> bytes:
        # Overestimate the required buffer by creating one the same size as the
        # input file as we should be able to safely assume the compressed file
        # will not be larger than the original file.
        output = create_string_buffer(size)

        # OodleLZ_Compress arguments:
        # 0: compressor     Which OodleLZ variant to use in compression
        # 1: rawBuf         Raw data to compress
        # 2: rawLen         Number of bytes in rawBuf to compress
        # 3: compBuf        Pointer to write compressed data to;
        #                   Should be at least $OodleLZ_GetCompressedBufferSizeNeeded
        # 4: level          OodleLZ_CompressionLevel controls how much CPU effort is put into maximizing
        #                   compression
        # 5: pOptions       (optional) Options; if NULL, $OodleLZ_CompressOptions_GetDefault is used
        # 6: dictionaryBase (optional) If not NULL, provides preceding data to prime the dictionary; must be
        #                   contiguous with rawBuf, the data between the pointers _dictionaryBase_ and _rawBuf
        #                   is used as the preconditioning data.  The exact same precondition must be passed
        #                   to encoder and decoder.
        # 7: lrm            (optional) Long range matcher
        # 8: scratchMem     (optional) Pointer to scratch memory
        # 9: scratchSize    (optional) Size of scratch memory (see $OodleLZ_GetCompressScratchMemBound)
        ret = self.handle.OodleLZ_Compress(
            9,  # compressor
            payload,  # rawBuf
            size,  # rawLen
            output,  # compBuf
            6,  # level
            None,  # pOptions
            None,  # dictionaryBase
            None,  # lrm
            None,  # scratchMem
            None,  # scratchSize
        )
        if ret != -1:
            return output.raw[:ret]
        else:
            raise OodleCompressionError(f"Error compressing {size} bytes: {ret}")

    def decompress(self, payload: bytes, size: int, output_size: int) -> bytes:
        """
        Decompress the payload using the given size.
        """
        output = create_string_buffer(output_size)

        # OodleLZ_Decompress arguments:
        # 0:  compBuf           Pointer to compressed data
        # 1:  compBufSize       Number of compressed bytes available (must be greater or equal to the number
        #                       consumed)
        # 2:  rawBuf            Pointer to output uncompressed data into
        # 3:  rawLen            Number of uncompressed bytes to output
        # 4:  fuzzSafe          (optional) Should the decode fail if it contains non-fuzz safe codecs?
        # 5:  checkCRC          (optional) If data could be corrupted and you want to know about it, pass
        #                       OodleLZ_CheckCRC_Yes
        # 6:  verbosity         (optional) If not OodleLZ_Verbosity_None, logs some info
        # 7:  decBufBase        (optional) If not NULL, provides preceding data to prime the dictionary; must
        #                       be contiguous with rawBuf, the data between the pointers _dictionaryBase_ and
        #                       _rawBuf_ is used as the preconditioning data. The exact same precondition must
        #                       be passed to encoder and decoder. The decBufBase must be a reset point.
        # 8:  decBufSize        (optional) Size of decode buffer starting at decBufBase, if 0, _rawLen_ is
        #                       assumed
        # 9:  fpCallback        (optional) OodleDecompressCallback to call incrementally as decode proceeds
        # 10: callbackUserData  (optional) Passed as userData to fpCallback
        # 11: decoderMemory     (optional) Pre-allocated memory for the Decoder, of size _decoderMemorySize_
        # 12: decoderMemorySize (optional) Size of the buffer at _decoderMemory_; must be at least
        #                       $OodleLZDecoder_MemorySizeNeeded bytes to be used
        # 13: threadPhase       (optional) For threaded decode; see $OodleLZ_About_ThreadPhasedDecode
        #                       (default OodleLZ_Decode_Unthreaded)
        ret = self.handle.OodleLZ_Decompress(
            c_char_p(payload),  # compBuf
            size,  # compBufSize
            output,  # rawBuf
            output_size,  # rawLen
            0,  # fuzzSafe
            0,  # checkCRC
            0,  # verbosity
            None,  # decBufBase
            None,  # decBufSize
            None,  # fpCallback
            None,  # callbackUserData
            None,  # decoderMemory
            None,  # decoderMemorySize
            3,  # threadPhase
        )

        # Make sure the result length matches the given output size
        if ret != output_size:
            raise OodleDecompressionError(
                "Decompression failed ret=0x{:x} output_size=0x{:x}".format(ret, output_size)
            )

        return output.raw
